#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""
raid崩壊時、
replicaiton設定をdisableにしないと
強制復帰後にreplication先のデータも消滅する可能性ある。

vfsミラーセット両方故障あるいは故障時、
replicaiton設定をdisableにしないと
replication先のデータも消滅する可能性ある。

vfsミラーセット両方抜けた状態でのNAS起動時、
内蔵mount時にreplicaiton設定をdisableにしないと
replication先のデータも消滅する可能性ある。

上記に対処するイベントハンドラ
"""

import re
import libevh.evhcmd as evhcmd
import libevh.util as util
import pal_info
import evh_config as config
from iobase.exception import NotFoundError
from iobase.xmldata import XmlConfNode
from ionas.request import Entry, EntryList
from product_conf.reqinfo import (Results, ReqTypes)


EVENTINFO_NAME = "eventinfo"
RAID_PATH = EVENTINFO_NAME + "/raid"
RAID_NAME_PATH = RAID_PATH + "/name"
DEVICE_PATH = EVENTINFO_NAME + "/device"
DEVICE_NAME_PATH = DEVICE_PATH + "/name"
VOLUME_PATH = EVENTINFO_NAME + "/volume"
VOLUME_NAME_PATH = VOLUME_PATH + "/name"
CATEGORY_ATTR = "category"
ACTION_ATTR = "action"

RAID_CATEGORY = "raid"
SATAD_CATEGORY = "satad"
USBD_CATEGORY = "usbd"
VOLUME_CATEGORY = "volume"
UPDATE_ACTION = "update"
FAIL_ACTION = "fail"
UNPLUG_ACTION = "unplug"
MOUNT_ACTION = "mount"

TRUE = "true"

REPLICATION_PATH = "replication"
REPLICATION_ENABLE_PATH = "replication/@enable"
STATUS_PATH = "raid/status"
SHARE_LIST_PATH = "sharelist/share/replication[@enable='true']/.."
PATH_NAME = "path"
FILESYSTEM_PATH = "volume/filesystem"
PARTITION_LIST_PATH = "volume/partitionlist/partition"

SHARES_NODE_STR = "<shares/>"

CRASH_STATUS = "crash"
DEGRADE_STATUS = "degrade"
VFSEXT_FS = "vfsext"
RAID_VOL = "md6"
MNT_POINT = "/mnt/sataraid1"
DISK1 = "satad1p6"
DISK2 = "satad2p6"
DISK3 = "satad3p6"
DISK4 = "satad4p6"
DISK5 = "satad5p6"
DISK6 = "satad6p6"

SATANUM_RE_COMPILE = re.compile("^satad([0-9]+)")
USBNUM_RE_COMPILE = re.compile("^usbd([0-9]+)")
USB_RE_COMPILE = re.compile("(^usbd[0-9]+)")


class CanNotConvError(Exception):
    pass


class ReplicationEscape(evhcmd.EvhCmd):

    def __add_entry(self, entrylist, share_node):

        shares_node = XmlConfNode(SHARES_NODE_STR)
        shares_node.append_child(share_node.get_clone())
        entrylist.append_entry(Entry(ReqTypes.UNSET, shares_node))

    def __set_disable(self, mnt_point):
        """
        共有設定ファイルSAVEと
        レプリケーションターゲットUNSETをする
        """

        repli_disable_entry_list = EntryList()

        share_config = self.send_entry_with_file(
            Entry(ReqTypes.LOAD),
            pal_info.PAL_CONFFILE_PATH,
            config.SHARES_CONF_FILE).get()

        for share_node in share_config.get_node_list(SHARE_LIST_PATH):
            if share_node.get_node(PATH_NAME).get_text().startswith(mnt_point):
                if share_node.get_value(REPLICATION_ENABLE_PATH) == TRUE:
                    share_node.get_node(REPLICATION_PATH).enable(False)
                    self.__add_entry(repli_disable_entry_list, share_node)

        if not repli_disable_entry_list.get_entry_list():
            return

        self.send_entry_with_file(
            Entry(ReqTypes.SAVE, share_config),
            pal_info.PAL_CONFFILE_PATH,
            config.SHARES_CONF_FILE)

        self.send_entrylist(
            repli_disable_entry_list,
            pal_info.PAL_REPLICATION_PATH)

    def __exec_disable(self):

        try:
            self.__set_disable(MNT_POINT)
        except NotFoundError:
            pass

    def __is_raid_crash(self):

        node = util.fal_raidctrl_query(RAID_VOL)
        if node.get_value(STATUS_PATH) == CRASH_STATUS:
            return True
        return False

    def __is_usb_disk(self, devname):

        if devname in config.devinfo.USB_HDD_DEV_LIST:
            return True

        return False

    def __num_conv(self, devname, re_compile):

        conv_devnum = re_compile.match(devname)

        if conv_devnum:
            return int(conv_devnum.group(1))
        raise CanNotConvError

    def __is_vfs_mirrorset_fail(self, entry, devname_xpath):

        node = util.fal_volctrl_query(config.INTERNAL_VOL)

        if node.get_value(FILESYSTEM_PATH) != VFSEXT_FS:
            return False

        devname = entry.get().get_value(devname_xpath)
        if (self.__is_usb_disk(devname) and
            not util.is_external_vfsext_setting(devname)):
            return False

        def is_mirror_fail(re_compile):
            exist_devnum_list = []

            for partname in node.get_value_list(PARTITION_LIST_PATH):
                try:
                    exist_devnum_list.append(
                        self.__num_conv(partname, re_compile))
                except CanNotConvError:
                    continue

            devname = entry.get().get_value(devname_xpath)
            try:
                fail_devnum = self.__num_conv(devname, re_compile)
            except CanNotConvError:
                return False

            # 偶数なら-1した奇数がミラーセット相手
            # 奇数なら+1した偶数がミラーセット相手
            if fail_devnum % 2:
                target_devnum = fail_devnum + 1
            else:
                target_devnum = fail_devnum - 1

            if target_devnum in exist_devnum_list:
                return False
            return True

        if (is_mirror_fail(SATANUM_RE_COMPILE) or
            is_mirror_fail(USBNUM_RE_COMPILE)):
            return True

    def __is_vfs_mirrorset_degrade_mount(self, entry):

        node = util.fal_volctrl_query(config.INTERNAL_VOL)

        if node.get_value(FILESYSTEM_PATH) != VFSEXT_FS:
            return False

        if entry.get().get_value(VOLUME_NAME_PATH) != config.INTERNAL_VOL:
            return False

        part_list = node.get_value_list(PARTITION_LIST_PATH)

        elements = config.EXT_RAID_TO_ELEMENT.values()[0].values()[0]

        sata_elms = [elm
                     for elm
                     in elements
                     if elm.startswith("satad")]
        usb_elms = [elm
                    for elm
                    in elements
                    if elm.startswith("usbd")]

        sata_degrade_info = []
        for elm in sata_elms:
            if not elm in part_list:
                sata_degrade_info.append(
                    config.EXT_RAID_DEV_INFO[elm]["mirrset"])
        for info in sata_degrade_info:
            num = len([other_info
                       for other_info
                       in sata_degrade_info
                       if other_info[0] == info[0]])
            if num >= 2:
                return True

        vfs_set_usbs = []
        for elm in usb_elms:
            usbdname = USB_RE_COMPILE.match(elm).group(1)
            if not util.is_external_vfsext_setting(usbdname):
                continue
            vfs_set_usbs.append(elm)

        for vfs_set_usb in vfs_set_usbs:
            if vfs_set_usb in part_list:
                break
        else:
            if vfs_set_usbs:
                return True

        return False

    def __disable(self, entry, devname_xpath):

        if self.__is_raid_crash():
            return

        if self.__is_vfs_mirrorset_fail(entry, devname_xpath):
            self.__exec_disable()

    def __is_target_ev(self, entry, cate, action):

        eventinfo = entry.get().get_node(EVENTINFO_NAME)
        if eventinfo.get_attr_value(CATEGORY_ATTR) == cate:
            if eventinfo.get_attr_value(ACTION_ATTR) == action:
                return True
        return False

    def __is_raidfail_ev(self, entry):

        if self.__is_target_ev(entry, RAID_CATEGORY, FAIL_ACTION):
            return True
        return False

    def __is_raidupdate_ev(self, entry):

        if self.__is_target_ev(entry, RAID_CATEGORY, UPDATE_ACTION):
            return True
        return False

    def __is_sataunplug_ev(self, entry):

        if self.__is_target_ev(entry, SATAD_CATEGORY, UNPLUG_ACTION):
            return True
        return False

    def __is_usbunplug_ev(self, entry):

        if self.__is_target_ev(entry, USBD_CATEGORY, UNPLUG_ACTION):
            return True
        return False

    def __is_volumemount_ev(self, entry):

        if self.__is_target_ev(entry, VOLUME_CATEGORY, MOUNT_ACTION):
            return True
        return False

    #
    # Request handlers
    #

    def __event_handler(self, entry):
        """Handle "event" request."""

        if self.__is_volumemount_ev(entry):
            if self.__is_vfs_mirrorset_degrade_mount(entry):
                self.__exec_disable()

        elif self.__is_raidupdate_ev(entry):
            if self.__is_raid_crash():
                self.__exec_disable()

        elif self.__is_raidfail_ev(entry):
            self.__disable(entry, RAID_NAME_PATH)

        elif self.__is_sataunplug_ev(entry) or self.__is_usbunplug_ev(entry):
            self.__disable(entry, DEVICE_NAME_PATH)

        entry.set_result(Results.OK)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        evhcmd.EvhCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.EVENT, self.__event_handler)

#
# Main
#

if __name__ == "__main__":
    ReplicationEscape(option=evhcmd.EvhCmdOption()).handle_request()
