#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Event Handler for unsetting replication."""

import os
import shutil
from ionas.request import Entry, EntryList
from iobase.xmldata import XmlConfNode
from iobase.exception import ExecError, NotFoundError
import evh_config as config
import libevh.evhcmd as evhcmd
import libevh.util as util
import fal_msginfo as msginfo
import pal_info
from product_conf.reqinfo import Results, ReqTypes

EVENTINFO_PATH = "eventinfo"
VOLUME_NAME_PATH = "volume/name"
SHARE_LIST_PATH = "sharelist/share/replication[@enable='true']/.."

CATEGORY_ATTR = "category"
ACTION_ATTR = "action"

VOLUME_CATEGORY = "volume"
SYSTEM_CATEGORY = "system"

UNMOUNT_ACTION = "unmount"
SHUTDOWN_ACTION = "shutdown"
REBOOT_ACTION = "reboot"

MNT_POINT = "mnt_point"
PATH_NAME = "path"
REPLICATION_FOLDER_NAME = "replication"

SHARES_NODE_FMT = "<shares>%s</shares>"


class EvhReplicationUnset(evhcmd.EvhCmd):
    """Event Handler class for unsetting replication."""

    #
    # Utility method
    #

    def __send_request(self, share_node_list):
        """Send unset request to pal_replication."""

        if not share_node_list:
            return

        entrylist = EntryList()

        for share_node in share_node_list:
            entrylist.append_entry(Entry(ReqTypes.UNSET,
                XmlConfNode(SHARES_NODE_FMT % str(share_node))))

        res_entrylist = self.send_entrylist(entrylist,
            pal_info.PAL_REPLICATION_PATH)

        for res_entry in res_entrylist.get_entry_list():
            if res_entry.get_result() != Results.OK:
                raise ExecError(pal_info.PAL_REPLICATION_PATH)

    def __get_share_node_list(self, root_path):
        """Get share node list."""

        share_config = self.send_entry_with_file(Entry(ReqTypes.LOAD),
            pal_info.PAL_CONFFILE_PATH, config.SHARES_CONF_FILE).get()

        share_node_list = []
        for share_node in share_config.get_node_list(SHARE_LIST_PATH):
            if share_node.get_node(PATH_NAME).get_text().startswith(root_path):
                share_node_list += [share_node]

        return share_node_list

    def __is_volume_unmount_event(self, category, action):
        """Check if event is volume unmount."""

        if category != VOLUME_CATEGORY:
            return False
        if action != UNMOUNT_ACTION:
            return False

        return True

    def __is_system_shutdown_event(self, category, action):
        """Check if event is system shutdown."""

        if category != SYSTEM_CATEGORY:
            return False
        if action != SHUTDOWN_ACTION:
            return False

        return True

    def __is_system_reboot_event(self, category, action):
        """Check if event is system reboot."""

        if category != SYSTEM_CATEGORY:
            return False
        if action != REBOOT_ACTION:
            return False

        return True

    def __is_valid_event(self, eventinfo_node):
        """Check if event is valid."""

        try:
            category = eventinfo_node.get_attr_value(CATEGORY_ATTR)
            action = eventinfo_node.get_attr_value(ACTION_ATTR)
        except:
            return False

        if self.__is_volume_unmount_event(category, action):
            # volume unmount
            volname_node = eventinfo_node.get_node(VOLUME_NAME_PATH)
            target_vol = volname_node.get_text()

            # internal volume only
            if target_vol == config.INTERNAL_VOL:
                return True
        elif self.__is_system_shutdown_event(category, action):
            # system shutdown
            return True
        elif self.__is_system_reboot_event(category, action):
            # system reboot
            return True

        return False

    def __try_unset_replication(self, root_path):
        """Unset all replication."""

        try:
            share_node_list = self.__get_share_node_list(root_path)
            if len(share_node_list) == 0:
                raise NotFoundError
        except NotFoundError:
            return

        try:
            self.__send_request(share_node_list)
            util.save_event_log(msginfo.Msg.MSG_REPLICATION_STOP)
        except:
            self.exception("failed to send unset request.")

    def __try_remove_replication_log_folder(self, root_path):
        """Remove replication log folder."""

        log_folder_path = os.path.join(root_path, REPLICATION_FOLDER_NAME)
        if not os.path.exists(log_folder_path):
            return

        try:
            shutil.rmtree(log_folder_path)
        except:
            self.exception("failed to remove replication log folder.")

    #
    # Request handlers
    #

    def __event_handler(self, entry):
        """Handle "event" request."""

        try:
            eventinfo_node = entry.get().get_node(EVENTINFO_PATH)
        except:
            self.exception("failed to get eventinfo node.")
            entry.set_result(Results.INVALID_XML)
            return entry

        if self.__is_valid_event(eventinfo_node):
            self.__try_unset_replication(
                config.VOL_INFO[config.INTERNAL_VOL][MNT_POINT])
            self.__try_remove_replication_log_folder(
                config.VOL_INFO[config.INTERNAL_VOL][MNT_POINT])

        entry.set_result(Results.OK)
        return entry

    def __init__(self):
        """Constructor."""

        evhcmd.EvhCmd.__init__(self, evhcmd.EvhCmdOption())
        self.set_handler(ReqTypes.EVENT, self.__event_handler)

#
# Main
#

if __name__ == "__main__":
    EvhReplicationUnset().handle_request()
